/*********************************************************/
/*       Clipping for lines and polygons with            */
/*        respect to a rectangular window                */
/*       Roy Williams   Copyright Caltech 1986           */
/*********************************************************/
#include "plotix.h"
#ifdef isMPW
#define __SEG__ PLOTIX
#include <:sys:config.h>
#include <:sys:plotix.h>
#else
#include "sys/config.h"
#include "sys/plotix.h"
#endif
extern char *_ex_malloc();
#ifndef DEVIND
#include <stdio.h>
#endif

static Coord xl, xr, yb, yt;    /* Window definition: 
                                   left, right, bottom, top */
struct point {
        Coord x, y;             /* Position of point */
        struct point *next;     /* Next point around the polygon */
};

static struct point *first, *last;
static int begin;
#if __STDC__                    /* Be nice to ourselves, use prototypes */
static int visible(struct point *this, int edge);
static struct point *palloc(void);
static struct point *clipit(struct point *first, int edge);
static int cross(struct point *A, struct point *B, int edge);
static void freelist(struct point *list);
static void insect(Coord *x, Coord *y, 
                   struct point *S, struct point *P, int edge);
#else                           /* No prototypes, but help out anyway */
static int visible();
static struct point *palloc();
static struct point *clipit();
static int cross();
static void freelist();
static void insect();
#endif                          /* __STDC__ */

/***********************************************************/
/*                 WINDOW DEFINITION                       */
/***********************************************************/

void _setclip(xlpass, ybpass, xrpass, ytpass)
Coord xlpass, ybpass, xrpass, ytpass;
{
        IND(softclip) = 1;
#ifndef isNCUBE                 /* NCUBE: compiler bug */
        xl = (xlpass < xrpass) ? xlpass : xrpass;
        yb = (ybpass < ytpass) ? ybpass : ytpass;
        xr = (xrpass > xlpass) ? xrpass : xlpass;
        yt = (ytpass > ybpass) ? ytpass : ybpass;
#else
        xl = xlpass; xr = xrpass;
        yb = ybpass; yt = ytpass;
#endif
}

void _endclip()
{
        IND(softclip) = 0;
}

/***********************************************************/
/*                    POINT CLIPPING                       */
/***********************************************************/

int _pointclip(x, y)
Coord x, y;
{
#ifndef isNCUBE         /* NCUBE: compiler bug */
        if(x < xl) return 0;
        if(x > xr) return 0;
        if(y < yb) return 0;
        if(y > yt) return 0;
#endif
        return 1;
}

/***********************************************************/
/*                    LINE CLIPPING                        */
/***********************************************************/

void _lineclip(x0, y0, x1, y1)
/*
        Clipping for lines.
        Author: Roy Williams, Caltech, December 1986.
        The code is copied from some pseudo-code in "Procedural Elements for 
        Computer Graphics", D. F. Rogers, McGraw-Hill, 1985, p 117.
*/
Coord x0, y0, x1, y1;
{
        struct point p0, p1, p, pp[2];
        int out0, out1, count;
        Coord m, x, y;
/*
        Set up a 4-bit string for each point
*/
        out0 = (x0 < xl) + (x0 > xr)*2 + (y0 < yb)*4 + (y0 > yt)*8;
        out1 = (x1 < xl) + (x1 > xr)*2 + (y1 < yb)*4 + (y1 > yt)*8;
/*
        Check for trivially invisible case
*/
        if(out0 & out1)
                return;
/*
        Set up drawing points pp[0] and pp[1]
*/
        p0.x = x0;              p0.y = y0;
        p1.x = x1;              p1.y = y1;
        pp[0] = p0;
        pp[1] = p1;
/*
        Check for totally visible line
*/
        if(!out0 && !out1)
                goto draw;
/*
        m set to infinite slope
*/
#ifdef INTCORD
#define LARGE 65536
#else
#define LARGE 1.e20
#endif
        m = LARGE;
/*
        Check for point 0 inside window
*/
        if(!out0){
                count = 1;
                pp[0] = p0;
                p = p1;
                goto leftcheck;
        }
/*
        Check for point 1 inside window
*/
        if(!out1){
                count = 1;
                pp[0] = p1;
                p = p0;
                goto leftcheck;
        }
/*
        Neither point inside window
*/
        count = -1;

again:
        if(count >= 0)
                pp[count] = p;
        if(++count > 1)
                goto draw;
        p = pp[count];

leftcheck:
        if(x0 - x1 == 0.0)
                goto topcheck;
        m = (y1 - y0)/(x1 - x0);
        if(xl < p.x)
                goto rightcheck;
        y = m*(xl - p.x) + p.y;
        if(y > yt || y < yb)
                goto rightcheck;
        p.y = y;
        p.x = xl;
        goto again;

rightcheck:
        if(xr > p.x)
                goto topcheck;
        y = m*(xr - p.x) + p.y;
        if(y > yt || y < yb)
                goto topcheck;
        p.y = y;
        p.x = xr;
        goto again;

topcheck:
        if(m == 0.0)
                goto again;
        if(yt > p.y)
                goto bottomcheck;
        x = (yt - p.y)/m + p.x;
        if(x < xl || x > xr)
                goto bottomcheck;
        p.x = x;
        p.y = yt;
        goto again;

bottomcheck:
        x = (yb - p.y)/m + p.x;
        if(x < xl || x > xr)
                return;
        p.x = x;
        p.y = yb;
        goto again;

draw:
        INVOKE(move)(pp[0].x, pp[0].y);
	IND(xold) = pp[0].x;
	IND(yold) = pp[0].y;
        INVOKE(cont)(pp[1].x, pp[1].y);
	IND(xold) = pp[1].x;
	IND(yold) = pp[1].y;
        return;
}

/*************************************************************/
/*                     POLYGON CLIPPING                      */
/*************************************************************/

/*
        Clipping for filled polygons.
        Author: Roy Williams, Caltech, December 1986
        Arguments
                N: Number of points to the polygon
                px, py: Arrays of x and y coordinates; may be anti- or clockwise.
                shade: Colour for the polygon
                edgestyle: 0 for plain, 1 for border
        We use the Sutherland-Hodgman Algorithm:
                "Reentrant Polygon Clipping", CACM 17 (1974) 32,
        which consists of clipping against the 4 edges of the rectangle in turn,
        producing a new polygon at each stage. Things get complicated because
        the intermediate and/or final clipped polygon may have more vertices
        than the original, so storage needs to be allocated, so we use a linked
        list for the points.
*/
void c_initpanel(shade, edgestyle)
int shade, edgestyle;
{
        struct point *palloc();
        INVOKE(initpanel)(shade, edgestyle);
        begin = 1;
}
void c_panelpoint(x, y)
/*
        Copy User information into linked list
*/
Coord x, y;
{
        struct point *this;
        this = palloc();
        this->x = x;
        this->y = y;
        if(!begin)
                last->next = this;
        else {
                first = this;
                begin = 0;
        }
        last = this;
	return;
}
void c_endpanel()
/*
        Now clip against each edge in turn
*/
{
        struct point *this, *new, *clipit();
        int edge;
/*
        Make the last connection to make the list circular
*/
        last->next = first;

#define         LEFT            0
#define         RIGHT           1
#define         BOTTOM          2
#define         TOP                     3
        for(edge=0; edge<4; edge++){
                new = clipit(first, edge);
                freelist(first);
                first = new;
                if(first == (struct point *)0)
                        return;
        }
/*
        Clipping done. Now just draw the result
*/
        this = first;
        do {
                INVOKE(panelpoint)(this->x, this->y);
        } while((this = this->next) != first);
        INVOKE(endpanel)();
        freelist(first);
}
static struct point *palloc()
/*
        Allocates memory for a 'struct vertex'
*/
{
        struct point *new;
        new = (struct point *)_ex_malloc(sizeof(struct point));
        if(new == (struct point *)0) {
#ifndef DEVIND
                fprintf(stderr, "Polygon Clipper: Malloc Failed\n");
#endif
                abort(-101);
        }
        return(new);
}
static void freelist(list)
/*
        Frees memory for a linked list of 'struct point's which is
        assumed to be a circular list
*/
struct point *list;
{
        struct point *this, *new;
        this = list;
        do {
                new = this->next;
                _ex_free(this);
                this = new;
        } while(this != list);
	return;
}
static struct point *clipit(first, edge)
/*
        Clips a polygon pointed-to by 'first' against the edge 'edge'.
        The new vertices are put into freshly mallocked space.
*/
struct point *first;
int edge;
{
        struct point *P, *S, *palloc(), *last, *new, *newfirst;
        int started;
        Coord xx, yy;
        started = 0;            /* Have we found any vertices for the new poly yet? */
        P = first;
        do {
/*
        Now we go around the old polygon. There are two types of new vertex:
        those which are intersections of the old polygon boundary with the
        clipping edge, and those which are visible points of the old polygon.
*/
                if(P != first && cross(S, P, edge)){    /* Intersection */
                        insect(&xx, &yy, S, P, edge);
                        if(started){
                                last = new;
                                new = palloc();
                                last->next = new;
                        }
                        else{
                                new = newfirst = palloc();
                                started = 1;
                        }
                        new->x = xx;
                        new->y = yy;
                }
                S = P;
                if(visible(S, edge)){                                   /* Visible Point */
                        if(started){
                                last = new;
                                new = palloc();
                                last->next = new;
                        }
                        else{
                                new = newfirst = palloc();
                                started = 1;
                        }
                        new->x = S->x;
                        new->y = S->y;
                }
        } while((P = P->next) != first);
/*
        Bit of an end-game here
*/
        if(started && cross(S, first, edge)){
                insect(&xx, &yy, S, first, edge);
                last = new;
                new = palloc();
                last->next = new;
                new->x = xx;
                new->y = yy;
        }
        if(!started)
                return((struct point *)0);              /* Completely invisible: return NULL */
        new->next = newfirst;
        return(newfirst);
}
static int visible(this, edge)
/*
        Is the point 'this' on the visible side of the edge?
*/
struct point *this;
int edge;
{
        switch (edge) {
                case LEFT:      return(this->x > xl);
                case RIGHT:     return(this->x < xr);
                case BOTTOM:return(this->y > yb);
                case TOP:       return(this->y < yt);
        }
}
static int cross(A, B, edge)
/*
        Does the line from A to B cross the edge ?
*/
struct point *A, *B;
int edge;
{
        return(visible(A, edge) ^ visible(B, edge));
}
static void insect(x, y, A, B, edge)
/*
        Puts into the pointers x and y the intersection point of the line AB
        with the edge. Nothing nasty should happen because the intersection
        has been approved by a successful call to 'cross'.
*/
Coord *x, *y;
struct point *A, *B;
int edge;
{
        switch (edge) {
                case LEFT:
                        *x = xl;
                        *y = (xl-A->x)*(B->y-A->y)/(B->x-A->x) + A->y;
                        return;
                case RIGHT:
                        *x = xr;
                        *y = (xr-A->x)*(B->y-A->y)/(B->x-A->x) + A->y;
                        return;
                case BOTTOM:
                        *y = yb;
                        *x = (yb - A->y)*(B->x-A->x)/(B->y-A->y) + A->x;
                        return;
                case TOP:
                        *y = yt;
                        *x = (yt - A->y)*(B->x-A->x)/(B->y-A->y) + A->x;
                        return;
        }
}
