/* 
   This program converts the measurements of the edge energy for a strings
   simulation at a given coupling (lambda) to a histogram of the number of
   configurations with a particular energy, which allows the construction 
   of the partition function (see Ferrenberg and Swendsen, Phys. Rev. Lett. 
   61, 2635 (1988)), and hence the energy and specific heat at any other 
   value of the coupling. 

   The data file to be analyzed is read from the input file (stdin), as
   are the values of the coupling for which the partition function is to be 
   calculated. These are specified by Kmin, Kmax and Kinc. Also read in
   are Emax and Emin, which give bounds on the values of the energy, 
   and Ebins for the number of bins.

   Errors are also given for all the points, using the formula taken
   from the paper of Huang, Moriarty, Myers and Potvin.

   Paul Coddington, June 8 1992.
*/

#include <stdio.h>
#include <string.h>
#include <math.h>

#define TRUE  1
#define FALSE 0

#define NOPS     15   /* number of operators measured */
#define OPERATOR  3   /* edge action */

#define NUMBINS 1000
/*
#define NUMBINS 500
*/

#define EPSILON 0.01  /* To avoid spurious truncations due to round-off error */
#define NUM_ERROR_BARS 0       /* For histogram plot */

float op[NOPS];
int hist[NUMBINS];  

char filename[128],fname[128],histname[128],dataname[128],datatype[128];
FILE *fopen(), *fpfs, *fphist, *infp;
double avenergy, avspheat, avpart;
double errener, varener, var, countsq;
double Kmin, Kmax, Kinc, Kprimed, diff;
double energy, Emax, Emin, Escale, Eoffset, Eround;
float lambda;
int i,j,k,n, x,y, nodes, cntenergy;
int Ebins, Kpoints;
int sumcounts, errint;
short ASCII, BINARY, errors;

/* Axis plots */
FILE *Efp, *Cfp, *HEfp, *FEfp;

main()
{
double sumenergy, sumenergysq, sumpart;
double part, boltz_factor;
int binenergy, histprint,histvalue;

   Ebins = NUMBINS;
   for (n=0;n<Ebins;n++) {
      hist[n] = 0;
   }

  /* Read in the data to construct the bins */

  fscanf(stdin, "%s", datatype);
  if (strncmp("ASCII",datatype,5) == 0) {
    ASCII = 1; BINARY = 0;
  }
  else if (strncmp("BINARY",datatype,6) == 0) {
    ASCII = 0; BINARY = 1;
  }
  else {
    perror("data type not specified");
    exit(-2);
  }
  
  fscanf(stdin,"%d",&nodes);   /* number of nodes in the mesh */
  fscanf(stdin,"%f",&lambda);  /* lambda value of data */

  fscanf(stdin,"%lf",&Kmin);   /* min value of K to be used */
  fscanf(stdin,"%lf",&Kmax);   /* max value of K to be used */
  fscanf(stdin,"%lf",&Kinc);   /* increment between K values */
    
  Kpoints = (int) (((Kmax - Kmin) / Kinc) + 1.0 + EPSILON);

  fscanf(stdin,"%lf",&Emin);   /* min value of E */
  fscanf(stdin,"%lf",&Emax);   /* max value of E */

  Escale = Emax - Emin;
  Eoffset = Escale / (2.0*(Ebins-1));  /* use midpoint value for bin energy */

  fscanf(stdin,"%s",fname); /* name of output file extension */

  while (fscanf(stdin,"%s",filename) != EOF) {

     if (find_pattern(filename, "HIST")) { 

	/* The histogram is already constructed, so read it in */
	printf("Reading from histogram file %s\n", filename);
	fflush(stdout);
        if (ASCII) read_hist(filename); 
        else if (BINARY) bread_hist(filename); 
	histprint = FALSE; 
        Escale = Emax - Emin;
        Eoffset = Escale / (2.0*(Ebins-1));  
     }
     else {
	/* Read in the raw data and histogram it */
	printf("Reading from data file %s\n", filename);
	fflush(stdout);
        if (ASCII) read_data(filename);    
        else if (BINARY) bread_data(filename);    
	histprint = TRUE;  /* write out the histogram */
     }

  } /* end of while loop reading from input file */

  /* Now calculate some stuff using the histogram */

  /* Open output datafile */
  sprintf(dataname, "FS.%s", fname); 
  fpfs = fopen(dataname,"w");

  fprintf(fpfs,"Nodes = %d\n", nodes);
  fprintf(fpfs,"Lambda = %12.8f\n",lambda);
  fprintf(fpfs,"Measurements = %d\n", cntenergy);
  fprintf(fpfs,"Energy range = %12.8f %12.8f\n", Emin, Emax);
  fprintf(fpfs,"Number of bins = %d\n", Ebins);
  fprintf(fpfs,"\n");
  fflush(fpfs);
  fprintf(fpfs,
    " lambda      Energy    Spec Heat  Partition Fn.\n");
  /* ........  ..........  ..........  ............ */
  fprintf(fpfs,"\n");

  /* Initialize Axis plots */
  init_axis_plots() ;

  if (NUM_ERROR_BARS == 0) errint = 1000000;
  else errint = Kpoints / NUM_ERROR_BARS;

  for (k=0; k<Kpoints; k++) {
       
     Kprimed = Kmin + (k * Kinc);
     diff = (Kprimed - lambda) * nodes;

     sumcounts = 0;
     sumpart = 0.;
     sumenergy = 0.; sumenergysq = 0.;

     /* Sum over bins to get averages */
     for (binenergy=0; binenergy<Ebins; binenergy++) {

	   histvalue = hist[binenergy];
	   sumcounts += histvalue;
	   if (histvalue != 0) {
             energy = (binenergy * Escale / (Ebins - 1)) + Emin + Eoffset;
	     boltz_factor = exp(-diff*energy);

	     /* Partition function */
	     part = histvalue * boltz_factor;
	     sumpart += part;

	     /* Energy and specific heat */
	     sumenergy += energy * part;
	     sumenergysq += energy * energy * part;

           }
     }

     if (sumcounts != cntenergy) {
	printf("ERROR! sumcounts = %d not equal to cntenergy = %d!\n",
		sumcounts, cntenergy); fflush(stdout);
     }
     avenergy = sumenergy / sumpart;
     avspheat = nodes * ((sumenergysq/sumpart)-(avenergy*avenergy));
     avspheat *= (Kprimed * Kprimed);
     avpart = sumpart / cntenergy;

     varener = 0.; 
     countsq = (double)(cntenergy * cntenergy);

     errors = ((k+1) % errint) == 0;
     if (errors) {
        /* Sum over bins again to get error estimates */
        for (binenergy=0; binenergy<Ebins; binenergy++) {
	   histvalue = hist[binenergy];
	   if (histvalue != 0) {
             energy = (binenergy * Escale / (Ebins - 1)) + Emin + Eoffset;
	     boltz_factor = exp(-diff*energy);
	     /* Factor in error expression */
	     part = histvalue / countsq;
	     part *= boltz_factor * boltz_factor;
	     /* Errors */
	     var = (energy - avenergy) * (energy - avenergy);
	     varener += var * part;
           }
        }
        varener /= (avpart * avpart);
        /* Work out autocorrelation factor */
	/*
        tau_ener = 0.5;
        errener = 2.0 * tau_ener * sqrt(varener);
	*/
	/* Assume data is uncorrelated */
        errener = sqrt(varener);
     } 

     /* Print out results */
     fprintf(fpfs,"%f  %10.6f  %10.6f    %E\n", 
             Kprimed,avenergy,avspheat,avpart);

     /* Axis plots */
     if (errors) {
        fprintf(Efp,"#e\n");
        fprintf(Efp,"%f %f %f\n", Kprimed,avenergy,errener);
        fprintf(Efp,"#e0\n#m 1\n");
     }
     else {
        fprintf(Efp,"%f %f\n", Kprimed,avenergy);
     }
     fprintf(Cfp,"%f %f\n", Kprimed,avspheat);

  } /* end of loop over number of K points */

     /* Axis plots  of histograms */
     for (binenergy=0; binenergy<Ebins; binenergy++) {
        histvalue = hist[binenergy];
        energy = (binenergy * Escale / (Ebins - 1)) + Emin;
	/* Histogram of energy */
        fprintf(HEfp,"%f %d\n", energy,histvalue);
	/* Free energy */
	if (histvalue !=0) {
          fprintf(FEfp,"%f %f\n", energy,-(log(histvalue*exp(-lambda*energy))));
	}
     }

  /* close output data file */
  fclose(fpfs)  ;

  /* Print out the histogram (if it does not already exist) */
  if (histprint) {
     sprintf(histname, "HIST.%s", fname);
     fphist = fopen(histname,"w");
     if (ASCII)        write_hist();
     else if (BINARY) bwrite_hist();
     fclose(fphist)  ;
  }

  exit(0);

} /* end of main */




/* SUBROUTINES */


read_hist(filename)
char *filename;
{
FILE *fphist;
int old_i,old_hist;
int new_nodes,new_Ebins;
float new_lambda;
   
   if ((fphist = fopen(filename,"r")) == NULL) {
      fprintf(stdout, "Histogram file %s not found\n", filename);
      fflush(stdout);
      exit(-1);
   }
   fscanf(fphist,"%d", &new_nodes);
   fscanf(fphist,"%d", &cntenergy);
   fscanf(fphist,"%f", &new_lambda);
   if (new_nodes != nodes) {
      printf("WARNING! nodes in histogram  = %d\n", new_nodes);
      printf("         nodes in input file = %d\n", nodes);
   }
   else {
      printf("Number of nodes = %d\n", new_nodes);
   }
   printf("Measurements = %d\n", cntenergy);
   fflush(stdout);
   if (new_lambda != lambda) {
      printf("WARNING! lambda in histogram  = %f\n", new_lambda);
      printf("         lambda in input file = %f\n", lambda);
   }
   else {
      printf("Lambda = %f\n", new_lambda);
   }
   fflush(stdout);
   fscanf(fphist,"%lf %lf", &Emin, &Emax);
   printf("Emin = %f  Emax = %f\n", Emin, Emax);
   fscanf(fphist,"%d", &new_Ebins);
   if (new_Ebins != Ebins) {
      printf("WARNING! Ebins in histogram = %d\n", new_Ebins);
      printf("         Ebins in program   = %d\n", Ebins);
   }
   else {
      printf("Number of bins = %d\n", Ebins);
   }
   fflush(stdout);
   while (fscanf(fphist,"%d %d\n", &old_i,&old_hist) != EOF) {
        hist[old_i] = old_hist;
   }
   fclose(fphist);
}


bread_hist(filename)
char *filename;
{
FILE *fphist;
int i;
short shorti;
int new_nodes,new_Ebins;
float new_lambda;

   if ((fphist = fopen(filename,"r")) == NULL) {
      fprintf(stdout, "Histogram file %s not found\n", filename);
      fflush(stdout);
      exit(-1);
   }
   fread(&new_nodes,sizeof(int),1,fphist);
   fread(&cntenergy,sizeof(int),1,fphist);
   fread(&new_lambda,sizeof(float),1,fphist);
   if (new_nodes != nodes) {
      printf("WARNING! nodes in histogram  = %d\n", new_nodes);
      printf("         nodes in input file = %d\n", nodes);
      fflush(stdout);
   }
   if (new_lambda != lambda) {
      printf("WARNING! lambda in histogram  = %f\n", new_lambda);
      printf("         lambda in input file = %f\n", lambda);
      fflush(stdout);
   }
   fread(&Emin,sizeof(double),1,fphist);
   fread(&Emax,sizeof(double),1,fphist);
   fread(&new_Ebins,sizeof(int),1,fphist);
   if (new_Ebins != Ebins) {
      printf("WARNING! Ebins in histogram = %d\n", new_Ebins);
      printf("         Ebins in program   = %d\n", Ebins);
      fflush(stdout);
   }
   while (fread(&shorti,sizeof(short),1,fphist) != 0) {
          fread(&hist[shorti],sizeof(int),1,fphist);
   }
   fclose(fphist);
}


write_hist()
{
int i;

   fprintf(fphist,"%d\n", nodes);
   fprintf(fphist,"%d\n", cntenergy);
   fprintf(fphist,"%f\n", lambda);
   fprintf(fphist,"%10.8f %10.8f\n", Emin, Emax);
   fprintf(fphist,"%d\n", Ebins);
   fprintf(fphist,"\n");
   for (i=0;i<Ebins;i++)
      if (hist[i] != 0) {
	 fprintf(fphist,"%d %d\n", i,hist[i]);
      }
}


bwrite_hist()
{
int i,histvalue;
short shorti;

   fwrite(&nodes,sizeof(int),1,fphist);
   fwrite(&cntenergy,sizeof(int),1,fphist);
   fwrite(&lambda,sizeof(float),1,fphist);
   fwrite(&Emin,sizeof(double),1,fphist);
   fwrite(&Emax,sizeof(double),1,fphist);
   fwrite(&Ebins,sizeof(int),1,fphist);
   for (i=0;i<Ebins;i++) {
      histvalue = hist[i];
      if (histvalue != 0) {
	 shorti = i;
         fwrite(&shorti,sizeof(short),1,fphist);
         fwrite(&histvalue,sizeof(int),1,fphist);
      }
   }
}


read_data(filename)
char *filename;
{
FILE *fp;
int nop,binenergy;

   if ((fp = fopen(filename,"r")) == NULL) {
      fprintf(stdout, "Data file %s not found\n", filename);
      fflush(stdout);
      exit(-1);
   }
   while(fscanf(fp, "%f", &op[0]) != EOF) {
      for (nop=1;nop<NOPS;nop++) {
         fscanf(fp, "%f", &op[nop]);
      }
      energy = op[OPERATOR];
      cntenergy++;
      energy = (energy - Emin) / Escale;   /* scale energy to lie in [0,1] */
      binenergy = (int) (energy * (Ebins - 1));
      if (binenergy < 0) {
         printf("energy = %f  binenergy = %d\n", energy, binenergy);
         fflush(stdout);
         binenergy = 0;
      }
      else if (binenergy >= Ebins) {
         printf("energy = %f  binenergy = %d\n", energy, binenergy);
         fflush(stdout);
         binenergy = Ebins - 1;
      }
      hist[binenergy]++;
   }
}


bread_data(filename)
char *filename;
{
FILE *fp;
int nop,binenergy;

   if ((fp = fopen(filename,"r")) == NULL) {
      fprintf(stdout, "Data file %s not found\n", filename);
      fflush(stdout);
      exit(-1);
   }
   while (fread(op,sizeof(float),NOPS,fp) != 0) {
      energy = op[OPERATOR];
      cntenergy++;
      energy = (energy - Emin) / Escale;   /* scale energy to lie in [0,1] */
      binenergy = (int) (energy * (Ebins - 1));
      if (binenergy < 0) {
         printf("energy = %f  binenergy = %d\n", energy, binenergy);
         fflush(stdout);
         binenergy = 0;
      }
      else if (binenergy >= Ebins) {
         printf("energy = %f  binenergy = %d\n", energy, binenergy);
         fflush(stdout);
         binenergy = Ebins - 1;
      }
      hist[binenergy]++;
   }
}


init_axis_plots()
{
   char filename[128];

   sprintf(filename, "E.%s.ax", fname);
   Efp = fopen(filename, "w");
   fprintf(Efp, "#g 1\n#lxs 1.4\n#lys 1.4\n#lts 1.4\n"); 
   fprintf(Efp, "#lty 1.07\n#ltx 0.5\n"); 
   fprintf(Efp, "#lt \"nodes=%d, \\gl =%f\"\n", nodes,lambda);
   fprintf(Efp, "#lx \"\\gl\"\n#ly \"Energy\"\n#lns 1.2\n#w 0.9\n#r .05\n#h 0.95\n#u .05\n#x %f %f\n",Kmin,Kmax);

   sprintf(filename, "C.%s.ax", fname);
   Cfp = fopen(filename, "w");
   fprintf(Cfp, "#g 1\n#lxs 1.4\n#lys 1.4\n#lts 1.4\n"); 
   fprintf(Cfp, "#lty 1.07\n#ltx 0.5\n"); 
   fprintf(Cfp, "#lt \"nodes=%d, \\gl =%f\"\n", nodes,lambda);
   fprintf(Cfp, "#lx \"\\gl\"\n#ly \"Specific heat\"\n#lns 1.2\n#w 0.9\n#r .05\n#h 0.95\n#u .05\n#x %f %f\n",Kmin,Kmax);

   sprintf(filename, "HE.%s.ax", fname);
   HEfp = fopen(filename, "w");
   fprintf(HEfp, "#g 1\n#lxs 1.4\n#lys 1.4\n#lts 1.4\n"); 
   fprintf(HEfp, "#lty 1.07\n#ltx 0.5\n"); 
   fprintf(HEfp, "#lt \"nodes=%d, \\gl =%f\"\n", nodes,lambda);
   fprintf(HEfp, "#lx \"Energy\"\n#ly \"Histogram\"\n#lns 1.2\n#w 0.9\n#r .05\n#h 0.95\n#u .05\n#x %f %f\n",Emin,Emax);

   sprintf(filename, "FE.%s.ax", fname);
   FEfp = fopen(filename, "w");
   fprintf(FEfp, "#g 1\n#lxs 1.4\n#lys 1.4\n#lts 1.4\n"); 
   fprintf(FEfp, "#lty 1.07\n#ltx 0.5\n"); 
   fprintf(FEfp, "#lt \"nodes=%d, \\gl =%f\"\n", nodes,lambda);
   fprintf(FEfp, "#lx \"Energy\"\n#ly \"Free energy\"\n#lns 1.2\n#w 0.9\n#r .05\n#h 0.95\n#u .05\n#x %f %f\n",Emin,Emax);
}


find_pattern(s,key)
char *s, *key;
{
char c, *ptr, *ptrkey;

  ptr = s; ptrkey = key;
  while ((c=*ptr++) != '\0') {
     if (*ptrkey==c) {
	if (*(++ptrkey)=='\0')
	   return 1;
     }
     else ptrkey = key;
  }
  return 0;
}

