/**********************************************/
/*    convect.c: convect a Gaussian around    */
/*               in a circle using explicit   */
/*               algorithm with no dissipation*/
/**********************************************/
#include "dime.h" 

struct user_node {
	double u;			/* Field to be convected */
	double upred;		/* Ditto predicted */
	double fx, fy;		/* Fluxes in x and y directions */
	double Q;			/* Convective part of du/dt */
	double cvarea;		/* Control volume area: sum of nbhd areas /3 */
};

struct user_bdy {
	double unused;
};

struct user_elmt {
	double fx, fy;			/* Average of nbhd node fluxes */
	double dx[3], dy[3];	/* Vectors pointing normal & out from edges */
	double area;			/* Elmt area */
	double refcrit;			/* Temp for adaptive refinement */
};

int step;				/* Total number of steps so far */
int need_resetmesh = 1;	/* Set to 1 after refine, relax etc */

double max_timestep;    /* Maximum timestep by CFL condition */
double timestep;        /* Actual timestep = safety factor * max timestep */
double time = 0;		/* Current simulation time */

#define GAUSSRAD 0.6    /* Radius to put the Gaussian lump */

user_main(mymenu, argc, argv)
MNUPTR mymenu;
int argc;
char **argv;
{
	int resetall();			/* Resets to initial condition */
	int solve();			/* Solves for the flow */
	int adaptrefine();		/* Adaptively refines a percentage of elmts */
	int draw_u();		    /* Contour the field */

	void convect_swops_func();	/* For DIME */
	void convect_refine_func();	/* For DIME */
#ifdef MESS_DEBUG
    int toggle_mess(), read_max_send();
#endif

	DIME_PREAMBLE;

/* Change the default handler functions */
	swops_func 	= convect_swops_func;
	refine_func = convect_refine_func;

/* Set up the menus */
	menu_add_function(mymenu, "Reset",		resetall);
	menu_add_function(mymenu, "Solve",		solve);
	menu_add_function(mymenu, "Adaptrefine",adaptrefine);
	menu_add_function(mymenu, "Draw_u",		draw_u);
#ifdef MESS_DEBUG
    menu_add_function(mymenu, "Debug_Mess",  toggle_mess);
    menu_add_function(mymenu, "Maxsend",  read_max_send);

#endif
}
#ifdef MESS_DEBUG
int dests[32];
int ison = 0;

read_max_send()
{
	int maxsend;
	printf("Max yto send?\n");
	get_int(&maxsend);
	set_maxsend(maxsend);
}
 
toggle_mess()
{
    int i;
    ison = 1-ison;
    for(i=0; i<nprocs; i++)
        dests[i] = 0;
}
 
mess_written(nbytes, dest)
int nbytes, dest;
{
    if(ison)
        dests[dest] += nbytes;
}
mess_flush()
{
    int i;
    if(!ison)
        return;
	fsingl(stdout);
    printf("%d n\n", nprocs);
    fmulti(stdout);
    for(i=0; i<nprocs; i++)
        printf("%d ", dests[i]);
    printf("t\n");
    fflush(stdout);
    fsingl(stdout);
    for(i=0; i<nprocs; i++)
        dests[i] = 0;
}
#endif

resetall()
/* 
	Reset  to a Gaussian at 3 o'clock
*/
{
	NODEPTR node;
	double x, y, rr, ww;

	printf("Gaussian Width? ");
	get_double(&ww);
	ww = ww*ww;

	FORALLNODES(node)
		if(B_NODE(node))
			node->user->u = 0;
		else {
			x = node->x - GAUSSRAD;
			y = node->y;
			rr = x*x + y*y;
			node->user->u = exp(-rr/ww);
		}
	NEXTNODE(node)
	time = 0;
}

resetmesh()
/*
	After refinement etc, some things need to be set up afresh:
		Element area
		Control volume areas
		Edge vectors
		Max timestep
*/
{
	NODEPTR node, node1, node2;
	ELMTPTR elmt;
	double area(), local_max_timestep;
	int i;

	FORALLNODES(node)
		node->user->cvarea = 0;
	NEXTNODE(node)

	max_timestep = 1.e12;
	FORALLELMTS(elmt)
		elmt->user->area = area(elmt);
/*
CFL condition is that time step should be < Maximum diameter of element
*/

		local_max_timestep = sqrt(elmt->user->area);
		if(local_max_timestep < max_timestep)
			max_timestep = local_max_timestep;

/*
(dx, dy) points normally out from element side
*/
		for(i=0; i<3; i++){
			node1 = elmt->neigh[(i+1)%3];
			node2 = elmt->neigh[(i+2)%3];
			elmt->user->dx[i] = 0.5*(node2->y - node1->y);
			elmt->user->dy[i] = 0.5*(node1->x - node2->x);
		}
/*
Control volume for node
*/
		for(i=0; i<3; i++)
			elmt->neigh[i]->user->cvarea += elmt->user->area/3;
	NEXTELMT(elmt)

	NODE_COMBINE(cvarea, fsum, 1);
	GLOBAL_COMBINE(max_timestep, fmin, 1);

	printf("Max CFL Timestep %f\n", max_timestep);
}

solve()
/*
	Ask for number of steps and timestep safety factor then 
	solve that number of iterations. Note timesteps are
	calculated only at the beginning before the first step
*/
{
	NODEPTR node;
	int fsum(), loops, loop, perplot;
	double beta;
	int step = 0;
/*
	Do we need to reset areas, edge vectors etc ?
*/
	GLOBAL_COMBINE(need_resetmesh, isum, 1);
	if(need_resetmesh){
		need_resetmesh = 0;
		resetmesh();
	}
/*
	Ask user for number of steps, safety factor
*/
	printf("Number of steps? "); fflush(stdout);
	get_int(&loops);

	printf("Timestep safety factor? "); fflush(stdout);
	get_double(&beta);
	timestep = max_timestep*beta;
	printf("Timestep %f\n", timestep);

	printf("Steps per plot\n");
	get_int(&perplot);

/*******************MAIN LOOP OF THE SOLVER**************/
	for(loop=0; loop<loops; loop++){

/* Copy current information into temp */
		FORALLNODES(node)
			node->user->upred = node->user->u;
		NEXTNODE(node)

/* 5 stage Runge Kutta scheme, dissipation evaluated at steps 1 & 2 */
		get_convect();
		update_upred(timestep/4);
		get_convect();
		update_upred(timestep/6);
		get_convect();
		update_upred(3*timestep/8);
		get_convect();
		update_upred(timestep/2);
		get_convect();
		update_upred(timestep);

/*
Copy prediction to actual field
*/
		FORALLNODES(node)
			if(!B_NODE(node))
				node->user->u = node->user->upred;
		NEXTNODE(node)

/*
Do a plot every so often
*/
		if(loop % perplot == perplot-1){
			disp_erase();
			draw_u();
		}
		time += timestep;
		printf("Step %d Time %f\n", step++, time);
	}
/*******************END MAIN LOOP OF THE SOLVER**********/
}

update_upred(deltat)
/*
	Use convective du/dt to update predicted
	u with fraction beta of a timestep
*/
double deltat;
{
	NODEPTR node;

/* Set -du/dt = residual = (convect - dissip) / control volume */
/* Now add in the residuals with the safety factor and local timestep */
	FORALLNODES(node)
		if(!B_NODE(node))
			node->user->upred = node->user->u 
				- deltat*node->user->Q/node->user->cvarea;
	NEXTNODE(node)
}

get_convect()
/*
	For a given information vector upred, calculate convective
	part of the residual in Q
*/
{
	NODEPTR node;
	ELMTPTR elmt;
	int i;

/* Get nodal fluxes */
	FORALLNODES(node)
		flux(node->x, node->y, node->user->upred, 
			&node->user->fx, &node->user->fy);
		node->user->fx /= 3.0;
		node->user->fy /= 3.0;
		node->user->Q = 0;
	NEXTNODE(node)

/* Average nodal fluxes into element fluxes */
	FORALLELMTS(elmt)
		elmt->user->fx = elmt->user->fy = 0.0;

		for(i=0; i<3; i++){
			node = elmt->neigh[i];
			elmt->user->fx += node->user->fx;
			elmt->user->fy += node->user->fy;
		}
	NEXTELMT(elmt)

/* Calculate flux across edges */
	FORALLELMTS(elmt)
		for(i=0; i<3; i++){
			node = elmt->neigh[i];
			node->user->Q += 
				elmt->user->fx * elmt->user->dx[i] +
				elmt->user->fy * elmt->user->dy[i];
		}
	NEXTELMT(elmt)

	NODE_COMBINE(Q, fsum, 1);
}

flux(x, y, u, fx, fy)
double x, y, u;
double *fx, *fy;
{
/*
	The flux shouldn't really use x and y but is here just to
	make the field rotate around
*/
	*fx = -y*u;
	*fy =  x*u;
}

draw_u()
{
	double u_func();
/*
Contour the field
*/
	slow_contor(u_func, 16, 1, -0.1, 1.1);
/*
Put in a marker to show where the lump should have got to
*/
	color(FOREGD);
	move(GAUSSRAD, 0.0);
	cont(0.0, 0.0);
	cont(GAUSSRAD*cos(time), GAUSSRAD*sin(time));
}

double u_func(node)
NODEPTR node;
/* for the contourer */
{
	return(node->user->u);
}

void
convect_swops_func(A, B, C, D, ABC, ACD, ABD, BCD)
NODEPTR A, B, C, D;
ELMTPTR ABC, ACD, ABD, BCD;
{
/* Don't need to do anything: almost everything is at the nodes */
	need_resetmesh = 1;
}

void
convect_refine_func(elmt, node, node1, node2, new, new1, new2)
ELMTPTR elmt, new1, new2;
NODEPTR node, node1, node2, new;
{
/*
	Need to interpolate the information vector.
*/
	new->user->u = 0.5*(node1->user->u + node2->user->u);
	need_resetmesh = 1;
}

adaptrefine()
/*
	Adapt the mesh according to the gradient of the pressure. User
	gives a percentage, and a threshold is found so that that
	percentage of elements will be refined.
*/
{
    ELMTPTR elmt;
    NODEPTR node;
    double a, gradx, grady, percent, ref_func();
    int i;
 
/*
	For each element, put the absolute gradient of the pressure
	into refcrit 
*/
    FORALLELMTS(elmt)
        gradx = grady = 0.0;
        a = sqrt(sqrt(elmt->user->area));
        for(i=0; i<3; i++){
            node = elmt->neigh[i];
            gradx += node->user->u * elmt->user->dx[i]/a;
            grady += node->user->u * elmt->user->dy[i]/a;
        }
        elmt->user->refcrit = sqrt(gradx*gradx + grady*grady);
    NEXTELMT(elmt)
 
/*
	Ask for the required percentage 
*/
	printf("Percentage? "); fflush(stdout);
	get_double(&percent);

	refine_fraction(ref_func, percent*0.01);
	swops();
}

double ref_func(elmt)
ELMTPTR elmt;
/* for the refiner */
{
	return(elmt->user->refcrit);
}

