/************************************************/
/*    laplace.c: solution of Laplace equation   */
/*               by Jacobi method, with         */
/*               Dirichlet Boundaries           */
/************************************************/
#include "dime.h"

struct user_node {
	double phi;            /* Solution of Laplace equation */
	double phinew;         /* Predicted solution */
	double diag;           /* Diagonal elements of stiffness matrix */
};

struct user_bdy	{
	double unused;
};

struct user_elmt {
	double stiff[3];        /* Local stiffness matrix */
	double refine_crit;		/* for refinement: magnitude of gradient of phi */
};

int setup_done = 0;

user_main(user_menu, argc, argv)
MNUPTR user_menu;
int argc;
char **argv;
{
	int solve();            /* Solve to convergence by Jacobi */
	int usr_contor();       /* Draw the solution */
	int refine_elmts();	    /* Refine according to gradient */

	void my_refine_func();

	DIME_PREAMBLE;

	menu_add_function(user_menu, "Solve", solve);
	menu_add_function(user_menu, "Contor", usr_contor);
	menu_add_function(user_menu, "Refine", refine_elmts);

	refine_func = my_refine_func;
}

solve()
{
	NODEPTR node;
	double jacobi();

	if(!setup_done){
		FORALLNODES(node)
			node->user->phi = 0.0;
			set_bdy_condition(node);
		NEXTNODE(node)
		setup_done = 1;
	}

/*
calculate stiffness matrix
*/
	stiffmat();
/*
keep doing Jacobi iteration until convergence
*/
	while(jacobi() > 0.001);
}

set_bdy_condition(node)
NODEPTR node;
{
/*
Boundary conditions are Dirichlet:
 Each boundary node must lie on a Curve or a Point.
 If the name of the Curve or Point begins with `s', phi = 1
 If the name of the Curve or Point begins with `t', phi = -1
 else phi = 0.02 * y-coordinate of node
*/
	CURVEPTR curve, c1, c2;
	double curvature();

	if(!B_NODE(node))
		return;

	if(curve = B_CURVE(node))
		node->user->phi = curvature(curve, node->b->s);
	else if((c1 = B_CURVE(node->b->clock)) && (c2 = B_CURVE(node->b->antik)))
		node->user->phi =0.5*(
			curvature(c1, node->b->clock->b->s) +
			curvature(c2, node->b->antik->b->s));
}

stiffmat()
{
/*
The matrix element fom node node1 to node node2 is the sum of the
cotangents of the angles opposite the edge connecting node1 to node2
*/
	NODEPTR node, node1, node2;
	ELMTPTR elmt;
	double a, area(), coeff, dx1, dy1, dx2, dy2;
	int i;

	FORALLNODES(node)
		node->user->diag = 0.0;
	NEXTNODE(node)

	FORALLELMTS(elmt)
		a = area(elmt);
		for(i=0; i<3; i++){
			node = elmt->neigh[i];
			node1 = elmt->neigh[(i+1)%3];
			node2 = elmt->neigh[(i+2)%3];
			dx1 = node1->x - node->x;	dy1 = node1->y - node->y;
			dx2 = node2->x - node->x;	dy2 = node2->y - node->y;
			elmt->user->stiff[i] = coeff = (dx1*dx2 + dy1*dy2)/a;
			node1->user->diag += coeff;
			node2->user->diag += coeff;
		}
	NEXTELMT(elmt)

	NODE_COMBINE(diag, fsum, 1)
}

double jacobi()
{
	int i, nnode = 0;
	NODEPTR node, node1, node2;
	ELMTPTR elmt;
	double deltaphi, err = 0;

	FORALLNODES(node)
		node->user->phinew = 0.0;
	NEXTNODE(node)

/*
Matrix multiply
*/
	FORALLELMTS(elmt)
		for(i=0; i<3; i++){
			node1 = elmt->neigh[(i+1)%3];
			node2 = elmt->neigh[(i+2)%3];
			node2->user->phinew += node1->user->phi * elmt->user->stiff[i];
			node1->user->phinew += node2->user->phi * elmt->user->stiff[i];
		}
	NEXTELMT(elmt)

	NODE_COMBINE(phinew, fsum, 1);

	FORALLNODES(node)
		if(!B_NODE(node)){
			nnode++;
			deltaphi = node->user->phinew / node->user->diag
				- node->user->phi;
			err += fabs(deltaphi);
			node->user->phi += deltaphi;
		}
	NEXTNODE(node)
	GLOBAL_COMBINE(err, fsum, 1);
	GLOBAL_COMBINE(nnode, isum, 1);
	printf("%f\n", err/nnode);

	return(err/nnode);
}

usr_contor()
{
	double contor_func();
	slow_contor(contor_func, 12, 1, 0.0, 0.0);
	return(1);
}

double contor_func(node)
NODEPTR node;
{
	return(node->user->phi);
}

refine_elmts()
{
	ELMTPTR elmt;
	NODEPTR node, node1, node2;
	double dx, dy, gradx, grady, elmtarea, elmtsize;
	double refcrit();
	int i;

/*
refinement criterion is the magnitude of the gradient of phi
times the diameter of the element
*/
	FORALLELMTS(elmt)
		gradx = grady = 0.0;
		for(i=0; i<3; i++){
			node = elmt->neigh[i];
			node1 = elmt->neigh[(i+1)%3];
			node2 = elmt->neigh[(i+2)%3];
			dx = node1->x - node2->x;
			dy = node1->y - node2->y;
			gradx += node->user->phi * dx;
			grady += node->user->phi * dy;
		}
		elmtarea = area(elmt);
		gradx /= elmtarea;
		grady /= elmtarea;

		elmtsize = sqrt(elmtarea);
		elmt->user->refine_crit = 
			(fabs(gradx) + fabs(grady)) * elmtarea;
	NEXTELMT(elmt)

	refine_fraction(refcrit, 0.25);
}

double refcrit(elmt)
ELMTPTR elmt;
{
	return(elmt->user->refine_crit);
}
	
void
my_refine_func(elmt, node, node1, node2, new, new1, new2)
ELMTPTR elmt, new1, new2;
NODEPTR node, node1, node2, new;
{
	new->user->phi = 0.5*(node1->user->phi + node2->user->phi);
	set_bdy_condition(new);
}
