/****************************************************/
/*    DIME: swops.c                                 */
/*    To do topological exchanges on the mesh       */
/*    to ensure diagonal dominance of the           */
/*    resulting stiffness matrix.                   */
/****************************************************/
/*        Roy Williams Copyright Caltech 1987       */
/****************************************************/
#define DIMECODE
#include "dime.h"

#define INVERSION		0        /* Possible returns from shall_we_swop */
#define WORSEN_ANGLES	1
#define IMPROVE_ANGLES	2

#define NOMESS			0		/* Do swops without messages sent */
#define MESSAGES		1		/* Do swops with messages */

#define NEWPROC			work_space.ei[0]	/* This is for send_elmts()
									and tells it where to send the elmts */
swops()
/*
	Carries out topological changes to the mesh. Each non-boundary edge
	has an elmt on each side: if the sum of the opposite angles is
	greater than 180 degress, then the edge is swopped to join the
	opposite nodes.

	This process is done internally to each processor, and the (x, y)
	of the nodes opposite processor-boundary edges are sent to all
	possible opposite processors, but only if that processor has a
	greater procnum. We don't want a parallel collision on this.

	When messages are received, they are checked to see if they are
	for us, and if so, if that processor edge wants swopping. If so,
	the whole elmt is marked for sending and sent with send_elmts().

	When the elmts are received, internal swops are done, but with no 
	messages, to make sure the new elmts are swopped as predicted
	by the other processor.

	The global number of swops done is combined, and if zero return else
	go around again.
*/
{
    int swops, count = 0;
	printf("Topo relax ");
	fflush(stdout);
	do {
		swops = 0;
#ifdef CUBIX
		if(mesh_is_distributed){
/*
	Do internal swops with messages.
*/
			swops += internal_swops(MESSAGES);
/*
	Send messages
*/
			if(crystal_router() == -1)
				abort(801);					/* CR failed in swops */
/*
	Process messages
*/
			get_swops_mess();
/*
	Send elmts to be given away
*/
			send_elmts();
		}
#endif
/*
	Do swops with no messages for the new ones
*/
		swops += internal_swops(NOMESS);
/*
	Get global number of swops done
*/
		if(mesh_is_distributed)
			combine(&swops, isum, sizeof(int), 1);

		printf("%d ", swops);
		fflush(stdout);
	} while(swops && count++ < 4);
	printf("\n");
}

internal_swops(flag)
int flag;
/*
	Do all internal swops. If flag == MESSAGES, then send the (x, y) of 
	the opposite node to a processor-boundary edge.
*/
{
    NODEPTR A, B, C, D;
    NODENEIGHPTR nstart, na, nan;
    ELMTPTR ABC, ACD;
    int abc_to_A, acd_to_A, swops, bedge, intedge, procedge;
	swops = 0;
	if(debug)
		fmulti(stdout);
	FORALLNODES(A)
nodeagain:
		na = nstart = A->n;
		do {
			nan = na->next;
			ABC = na->elmt;             abc_to_A = indexc(ABC, A);
			ACD = nan->elmt;            acd_to_A = indexc(ACD, A);
			C = ABC->neigh[(abc_to_A + 2) % 3];
			B = ABC->neigh[(abc_to_A + 1) % 3];
/*
    An edge has the property bedge if it is part of the boundary,
    either a BOUNDARY or CUT boundary. 
	intedge means a true internal edge.
*/
			bedge = (A->b != (BDYPTR)NULL) && (C->b != (BDYPTR)NULL) && (A->b->clock == C);
			intedge = !bedge && (long)A < (long)C;
			procedge = bedge && A->b->oppalias != (ALIASPTR)NULL 
				&& !(A->b->oppalias->proc & HIBITS);

			if(procedge && flag == MESSAGES)
				mess_swop(A, B->x, B->y);
			if(intedge){
				D = ACD->neigh[(acd_to_A + 2) % 3];
				if(shall_we_swop(A, B, C, D) == IMPROVE_ANGLES
						&& !connected(B, D)){
					swop(A, B, C, D, ABC, ACD);
					swops++;
					goto nodeagain;
				}
			}
		} while((na = nan) != nstart);
	NEXTNODE(A)
	if(debug){
		if(mesh_is_distributed)
			usendplot();
		else
			sendplot();
		fsingl(stdout);
	}
	return(swops);
}

swop(A, B, C, D, ABC, ACD)
/*
    Carries out a swop on the quarilateral ABCD (anticlockwise), where AC
    is at present connected and on return BD will be. 
*/
NODEPTR A, B, C, D;
ELMTPTR ABC, ACD;
{
	NODENEIGHPTR remove(), nodeneigh_alloc();
	ELMTPTR ABD, BCD, elmt_alloc();
	ABD = elmt_alloc();
	BCD = elmt_alloc();
	ABD->neigh[0] = A;	ABD->neigh[1] = B;	ABD->neigh[2] = D;
	BCD->neigh[0] = B;	BCD->neigh[1] = C;	BCD->neigh[2] = D;
	insert(A, ACD, nodeneigh_alloc(), ABD);
	insert(B, ABC, nodeneigh_alloc(), BCD);
	insert(B, BCD, nodeneigh_alloc(), ABD);
	insert(C, ABC, nodeneigh_alloc(), BCD);
	insert(D, ACD, nodeneigh_alloc(), ABD);
	insert(D, ABD, nodeneigh_alloc(), BCD);
	if(swops_func != NULL)
		(*swops_func)(A, B, C, D, ABC, ACD, ABD, BCD);
	sfree(nntype, remove(A, ABC));
	sfree(nntype, remove(A, ACD));
	sfree(nntype, remove(B, ABC));
	sfree(nntype, remove(C, ABC));
	sfree(nntype, remove(C, ACD));
	sfree(nntype, remove(D, ACD));
	freeelmt(ABC);
	freeelmt(ACD);
	get_bdy_nn(A);
	get_bdy_nn(B);
	get_bdy_nn(C);
	get_bdy_nn(D);
    if(debug){
        color(0);draw(A, C);
        color(1);draw(B, D);
    }
}

mess_swop(node, x, y)
NODEPTR node;
double x, y;
/*
	Sends the message (x, y) to all processors which are aliased both to
	node and nodenext. It is assumed that node->b->next is nodenext.
*/
{
	char *mailbag;
	if(node->b == (BDYPTR)NULL || node->b->oppalias == (ALIASPTR)NULL)
		abort(802);					/* Non-boundary in mess_swop */
	if(node->b->oppalias->proc > procnum)
		return;
	mailbag = buffer;
	xpacket(mailbag, NODEPTR ) = node->b->oppalias->node;
	xpacket(mailbag, double) = x;
	xpacket(mailbag, double) = y;
	if(add_to_out_buf(buffer, (int)(mailbag-buffer), 1, 
			&(node->b->oppalias->proc)) == -1)
		abort(803);					/* AOB failed in mess_swop */
}

get_swops_mess()
/*
	Analyse all the messages received. First check that the given nodes
	are in downstream order for us, ie that the message is really directed 
	here, then use the node template D to put the (x, y) into, and ask
	if it should be swopped. If so, load up that cell to be shipped off
	by send_elmts().
*/
{
	int source, nbytes, i;
	ELMTPTR elmt;
	NODENEIGHPTR n, nstart;
	NODEPTR A, B, C;
	struct node D;
	char *mailbag;
/*
	First make workspace so that no elmt is moved anywhere.
*/
	FORALLELMTS(elmt)
		elmt->NEWPROC = procnum;
	NEXTELMT(elmt)
/*
	Read Messages
*/
	while(source = ANY, get_message(buffer, &source, &nbytes) == 0){
		mailbag = buffer;
		A =		xpacket(mailbag, NODEPTR );
		D.x =	xpacket(mailbag, double);
		D.y =	xpacket(mailbag, double);
		if(A->b == (BDYPTR)NULL)
			abort(804);					/* Non-boundary in get_swops_mess */
		C =		A->b->clock;
		n = nstart = A->n;
		do {
			elmt = n->elmt;
			i = indexc(elmt, A);
			if(elmt->neigh[(i+2)%3] == C){
				B = elmt->neigh[(i+1)%3];
				if(shall_we_swop(A, B, C, &D) == IMPROVE_ANGLES){
					elmt->NEWPROC = source;
					break;
				}
			}
		} while((n = n->next) != nstart);
	}
}

connected(n1, n2)
NODEPTR n1, n2;
{
	ELMTPTR elmt;
	int i;
	if(n1 == n2)
		return(1);

	FORALLNEIGH(n1, elmt)
		for(i=0; i<3; i++)
			if(elmt->neigh[i] == n2)
				return(1);
	NEXTNEIGH(n1, elmt)
	return(0);
}

default_shall_we_swop(A, B, C, D)
/*
    Decides on the status of a proposed swop: The reply may be
        INVERSION: This swop will create a negative-area triangle.
        WORSEN_ANGLES: The swop is acceptable, but does not improve the mesh.
        IMPROVE_ANGLES: The swop improves the mesh.
*/
NODEPTR A, B, C, D;
{
    double a, c;
    double BAx, BAy, DAx, DAy, BCx, BCy, DCx, DCy;
    BAx = B->x - A->x;            BAy = B->y - A->y;
    DAx = D->x - A->x;            DAy = D->y - A->y;
    BCx = B->x - C->x;            BCy = B->y - C->y;
    DCx = D->x - C->x;            DCy = D->y - C->y;
    a = BAx*DAy - BAy*DAx;
    c = DCx*BCy - DCy*BCx;
    if(a < 0.0 || c < 0.0)
        return(INVERSION);
    if(a*(DCx*BCx + DCy*BCy) + c*(DAx*BAx + DAy*BAy) <= 0.0)
        return(WORSEN_ANGLES);
    else
        return(IMPROVE_ANGLES);
}
