/****************************************************/
/*    DIME: refine.c                                */
/*    To locally refine the mesh.                   */
/****************************************************/
/*        Roy Williams Copyright Caltech 1987       */
/****************************************************/
#define DIMECODE
#include "dime.h"

#define REFINE		work_space.ei[0]	/* Elmt opposite nonconforming node */
#define NCNODE		work_space.ni[0]	/* Nonconforming node */
#define OPP_NC_PT	work_space.ba[0]	/* Address in other proc of NC node */
#define OPP_NC_PROC	work_space.bi[1]	/* Procnum of other proc */

set_refine(elmt, flag)
ELMTPTR elmt;
int flag;
{
	elmt->REFINE = flag;
}

refine()
/*
	Carries out the algorithm described in:

		"Design and Data Structure of Fully Adaptive Multigrid,
		Finite-Element Software", Maria-Cecilia Rivara,
		ACM Trans in Math Software, Vol 10 (1984) 242-264.

	for local refinement of a triangular mesh.

	It is assumed that the user has filled workspace.ei[0] for each
	element with truthvalue as to whether this element is to be refined.

	Each chosen element is divided in two by creating a node on the longest
	side. The new node is nonconforming because of the T shape.
	Whenever a processor-boundary edge is divided, a message is
	queued to the other side.

	Messages sent. Addresses of nodes in other processors are squirreled 
	away in the boundary-structure workspace.  All elements opposite 
	nonconforming nodes are marked for refinement, and the number of these
	counted.

	Get global number to be refined by combine(), and return if zero,
	else back to start and refine these.
*/
{
	ELMTPTR elmt, elmtnext, elmt1, elmt2, elmt_alloc();
	NODEPTR node, node1, node2, new, node_alloc(), ncpoint();
	NODENEIGHPTR nodeneigh_alloc(), remove();
	ALIASPTR ali, alias_alloc();
	BDYPTR b_alloc();
	int i, create, toberefined, ref, intedge, procedge;
	printf("Refine ");
	fflush(stdout);
/*
	Zero the work spaces
*/
	FORALLNODES(node)
		node->NCNODE = 0;
		if(node->b != (BDYPTR)NULL)
			node->b->OPP_NC_PT = (NODEPTR)NULL;
	NEXTNODE(node)

	while(1){
/*
	Get global number to be refined. If zero we are done.
*/
		toberefined = 0;
		FORALLELMTS(elmt)
			toberefined += (elmt->REFINE != 0);
		NEXTELMT(elmt)
		combine(&toberefined, isum, sizeof(int), 1);
		fsingl(stdout);
		if(toberefined == 0){
			printf("\n"); fflush(stdout);
			return;
		}
		printf("%d ", toberefined); fflush(stdout);
		if(debug){
			color(1);
			sendplot();
		}
		fmulti(stdout);
/*
	This is the refinement of those elmts marked.
*/
		FORALLELMTS(elmt)
			if(elmt->REFINE == 0)
				continue;
			node = elmt->neigh[i = longest(elmt)];
			node1 = elmt->neigh[(i+1)%3];
			node2 = elmt->neigh[(i+2)%3];

			new = ncpoint(elmt, node2, node1);
			create = (new == (NODEPTR)NULL);
			procedge = 0;
			if(create){
				new = node_alloc();
				intedge = (node2->b == (BDYPTR)NULL || node2->b->clock != node1);
				if(!intedge){
					new->b = b_alloc();
					new->b->OPP_NC_PT = (NODEPTR)NULL;
					procedge = !B_EDGE(node2);
					node2->b->clock = new;
					node1->b->antik = new;
					new->b->clock = node1;
					new->b->antik = node2;
					if(!procedge)
						new->b->is_on_bdy |= 1;
					if(node2->b->oppalias != (ALIASPTR)NULL 
							&& node2->b->oppalias->proc&HIBITS)
						new->b->is_on_bdy |= 2;
				}
				new->x = 0.5*(node1->x + node2->x);
				new->y = 0.5*(node1->y + node2->y);
				new->NCNODE = (intedge || procedge);
			} else
				new->NCNODE = 0;

			elmt1 = elmt_alloc();
			elmt2 = elmt_alloc();
			elmt1->REFINE = elmt2->REFINE = 0;
			elmt1->neigh[0] = elmt2->neigh[0] = new;
			elmt1->neigh[1] = node;
			elmt1->neigh[2] = node1;
			elmt2->neigh[1] = node2;
			elmt2->neigh[2] = node;
			add_neigh(elmt1, new);
			insert(new, elmt1, nodeneigh_alloc(), elmt2);
			insert(node, elmt, nodeneigh_alloc(), elmt1);
			insert(node, elmt1, nodeneigh_alloc(), elmt2);
			insert(node1, elmt, nodeneigh_alloc(), elmt1);
			insert(node2, elmt, nodeneigh_alloc(), elmt2);
/*
	Here the user can put stuff in his part of the new structures using the
	old structures as a guide.
*/
			if(refine_func != NULL)
				(*refine_func)(elmt, node, node1, node2, new, elmt1, elmt2);
			if(debug)
				draw(node, new);
			sfree(nntype, remove(node, elmt));
			sfree(nntype, remove(node1, elmt));
			sfree(nntype, remove(node2, elmt));
			get_bdy_nn(node);
			get_bdy_nn(node1);
			get_bdy_nn(node2);
			if(!create || !intedge)
				sortout(new);
			if(procedge){
/*
	Have bisected a processor-boundary edge: send a message across
*/
				mess_refine(node2, new);
				new->b->oppalias = alias_alloc();
				new->b->alias = ali = alias_alloc();
				ali->next = ali;
				if(node2->b->OPP_NC_PT != (NODEPTR)NULL){
					new->b->oppalias->node = node2->b->oppalias->node;
					new->b->oppalias->proc = node2->b->oppalias->proc;
					node2->b->oppalias->node = ali->node = node2->b->OPP_NC_PT;
					node2->b->oppalias->proc = ali->proc = 
													node2->b->OPP_NC_PROC;
					node2->b->OPP_NC_PT = (NODEPTR)NULL;
				}
			}
		NEXTELMT(elmt)
/*
	Now do the tricky business of deleting entries from a linked list
	while cycling through it
*/
		if(the_elmts != (ELMTPTR)NULL){
			elmtnext = the_elmts->next;
			do {
				elmt = elmtnext->last;
				if(elmt->REFINE)
					freeelmt(elmt);
			} while((elmtnext = elmtnext->next) != the_elmts);

			if(the_elmts->last->REFINE)
				freeelmt(elmt);
		}
/*
	Refining done. Send queued messages.
*/
		if(debug){
			if(mesh_is_distributed)
				usendplot();
			else
				sendplot();
		}

		if(crystal_router() == -1)
			abort(501);					/* CR failed in refine */
/*
	Messages sent. Deal with them.
*/
		while(get_refine_message());
/*
	Now decide on which elmts to be refined for the next round. An elmt
	is either opposite an NC point or a message came from another processor.
*/
		FORALLELMTS(elmt)
			ref=0;
			for(i=0; i<3; i++){
				node1 = elmt->neigh[(i+1)%3];
				node2 = elmt->neigh[(i+2)%3];
				ref |= (ncpoint(elmt, elmt->neigh[i], node2) != (NODEPTR)NULL);
				if(node2->b != (BDYPTR)NULL && node2->b->clock == node1)
					ref |= (node2->b->OPP_NC_PT != (NODEPTR)NULL);
			}
			elmt->REFINE = ref;
		NEXTELMT(elmt)
	}
}

longest(elmt)
ELMTPTR elmt;
/*
	Finds the longest side of an elmt and returns the index of the node
	opposite. If any node of the cell is an NC node, that index is returned.
*/
{
	double dmax = 0.0, d;
	NODEPTR n1, n2;
	int i, imax;
	for(i=0; i<3; i++){
		if(elmt->neigh[i]->NCNODE)
			return(i);
		n1 = elmt->neigh[(i+1)%3];
		n2 = elmt->neigh[(i+2)%3];
		d = (*sqdist)(n1, n2);
		if(d > dmax){
			dmax = d;
			imax = i;
		}
	}
	return(imax);
}

NODEPTR ncpoint(elmt, node, node2)
ELMTPTR elmt;
NODEPTR node, node2;
/*
	Finds an NC node opposite an elmt if there is one, else returns NULL.
*/
{
	NODENEIGHPTR indexnn();
	ELMTPTR elmtn, elmtnn;
	NODEPTR noden, nodenn;
	if(node->b != (BDYPTR)NULL && node->b->clock == node2)
		return((NODEPTR)NULL);
	elmtn = (indexnn(node, elmt))->next->elmt;
	noden = elmtn->neigh[(indexc(elmtn, node)+1)%3];
	elmtnn = (indexnn(noden, elmtn))->next->elmt;
	nodenn = elmtnn->neigh[(indexc(elmtnn, noden)+1)%3];
	return((noden->NCNODE && nodenn == node2) ? noden:(NODEPTR)NULL);
}

mess_refine(node, new)
NODEPTR node, new;
/*
	Sends messages to all likely processors that a NC node 'new' has 
	been created between 'node' and 'nodenext'.
*/
{
	char *mailbag;
	int dest;
	if(node->b == (BDYPTR)NULL || node->b->oppalias == (ALIASPTR)NULL)
		abort(502);						/* Non-boundary node in refine */
	mailbag = buffer;
	xpacket(mailbag, NODEPTR ) = node->b->oppalias->node;
	xpacket(mailbag, NODEPTR ) = new;

	if(node->b->oppalias->proc & HIBITS)
		xpacket(mailbag, int) =  procnum | HIBITS;
	else
		xpacket(mailbag, int) =  procnum;

	dest = (node->b->oppalias->proc) & LOWBITS;
	if(add_to_out_buf(buffer, (int)(mailbag-buffer), 1, &dest) == -1)
		abort(503);						/* AOB failed in refine */
}

get_refine_message()
/*
	Reads messages from the other processors. First check if the message
	is really directed to us, then either:
		if an NC node is already on this edge, hook it up with the
		address in the other processor,
	or:
		if no NC node, put the foreign address and procnum in the
		boundary node workspace upstream from where our new node will be.
*/
{
	int source, nbytes;
	NODEPTR resident, new, next;
	char *mailbag;
	if(source = ANY, get_message(buffer, &source, &nbytes) != 0)
		return(0);
	mailbag = buffer;
	resident =	xpacket(mailbag, NODEPTR );
	new =		xpacket(mailbag, NODEPTR );
	source =    xpacket(mailbag, int);
	next = resident->b->clock;
	if(next->NCNODE){
		next->b->alias->node = new;
		next->b->alias->proc = source;
		next->b->oppalias->node = resident->b->oppalias->node;
		next->b->oppalias->proc = resident->b->oppalias->proc;
		resident->b->oppalias->node = new;
		resident->b->oppalias->proc = source;
	} else {
		resident->b->OPP_NC_PT = new;
		resident->b->OPP_NC_PROC = source;
	}
	return(1);
}
