/****************************************************/
/*    DIME: alloc.c                                 */
/*    This code used to allocate and free           */
/*    the various structures used by DIME.          */
/****************************************************/
/*        Roy Williams Copyright Caltech 1987       */
/****************************************************/

/* Also contains code which used to be in smalloc.c */

#define DIMECODE
#include "dime.h"

int ntype, untype, etype, uetype, btype, ubtype, atype, nntype;

NODEPTR node_alloc()
/*
Allocate memory for a new node, with nneigh neighnode structures attached
in a list. The new node is also put in the circular list of all nodes.
*/
{
	NODEPTR this;
	int in;
	static int first = 1;
	if(first){
		ntype = init_smalloc(sizeof(_NODE));
		untype = init_smalloc(uns);
		first = 0;
#ifdef DEBUG
	fprintf(stdsys, "node type %d size %d\n", ntype, sizeof(_NODE));
	fprintf(stdsys, "node user type %d size %d\n", untype, uns);
#endif
	}
	this = (NODEPTR)smalloc(ntype);
	if(this == (NODEPTR)NULL)
		abort(100);		/* malloc failed for node */
#ifdef DEBUG
printf(">>> node = %x, %d rem 8 = %d\n", this, this, ((long)this)%8);
#endif
	this->b = (BDYPTR)NULL;
	this->n = (NODENEIGHPTR)NULL;
	if(the_nodes == (NODEPTR)NULL)
		the_nodes = this->last = this->next = this;
	else {
		this->next = the_nodes;
		this->last = the_nodes->last;
		the_nodes->last->next = this;
		the_nodes->last = this;
	}

	if(uns > 0)
		if((this->user = smalloc(untype)) == (char *)NULL)
			abort(101);	/* malloc failed for user_node */
	return(this);
}

NODENEIGHPTR nodeneigh_alloc()
/*
	Allocate space for the structure which points to a elmt from a node.
*/
{
	NODENEIGHPTR n;
	static int first = 1;
	if(first){
		nntype = init_smalloc(sizeof(NODENEIGH));
		first = 0;
#ifdef DEBUG
	fprintf(stdsys, "nodeneigh type %d size %d\n", nntype, sizeof(NODENEIGH));
#endif
	}
	n = (NODENEIGHPTR )smalloc(nntype);
	if(n == (NODENEIGHPTR)NULL)
		abort(102);		/* malloc failed for nodeneigh */
	return(n);
}

BDYPTR b_alloc()
/*
	Allocate space for the structure which points to a elmt from a node.
*/
{
	BDYPTR n;
	static int first = 1;
	if(first){
		btype = init_smalloc(sizeof(_BDY));
		ubtype = init_smalloc(ubs);
		first = 0;
#ifdef DEBUG
	fprintf(stdsys, "bdy type %d size %d\n", btype, sizeof(_BDY));
	fprintf(stdsys, "bdy user type %d size %d\n", ubtype, ubs);
#endif
	}
	n = (BDYPTR )smalloc(btype);
	if(n == (BDYPTR)NULL)
		abort(103);		/* malloc failed for bdy */
	n->alias = (ALIASPTR)NULL;
	n->oppalias = (ALIASPTR)NULL;
	n->is_on_bdy = 0;
	if(ubs > 0)
		if((n->user = smalloc(ubtype)) == (char *)NULL)
			abort(104);	/* malloc failed for bdy_user */
	return(n);
}

ELMTPTR elmt_alloc()
/*
	Allocate space for a elmt, and put it in the circular list of all elmts
*/
{
	ELMTPTR this;
	static int first = 1;
	if(first){
		etype = init_smalloc(sizeof(_ELMT));
		uetype = init_smalloc(ues);
		first = 0;
#ifdef DEBUG
	fprintf(stdsys, "elmt type %d size %d\n", etype, sizeof(_ELMT));
	fprintf(stdsys, "elmt user type %d size %d\n", uetype, ues);
#endif
	}
	this = (ELMTPTR )smalloc(etype);
	if(this == (ELMTPTR)NULL)
		abort(105);		/* malloc failed for elmt */
	if(the_elmts == (ELMTPTR)NULL)
		the_elmts = this->last = this->next = this;
	else {
		this->next = the_elmts;
		this->last = the_elmts->last;
		the_elmts->last->next = this;
		the_elmts->last = this;
	}
	if(ues > 0)
		if((this->user = smalloc(uetype)) == (char *)NULL)
			abort(106);	/* malloc failed for user_elmt */
	return(this);
}

ALIASPTR alias_alloc()
/*
	Allocate space for an alias-structure
*/
{
	ALIASPTR new;
	static int first = 1;
	if(first){
		atype = init_smalloc(sizeof(ALIAS));
		first = 0;
#ifdef DEBUG
	fprintf(stdsys, "alias type %d size %d\n", atype, sizeof(ALIAS));
#endif
	}
	new = (ALIASPTR )smalloc(atype);
	if(new == (ALIASPTR)NULL)
		abort(107);		/* malloc failed for alias */
	return(new);
}

freemesh()
/*
	Frees everything associated with the current mesh and sets the_nodes and
	the_elmts to NULL.
*/
{
	NODEPTR node;
	ELMTPTR elmt;
	if((node = the_nodes) != (NODEPTR)NULL){
		node = node->next;
		do {
			freenode(node);
		} while((node = node->next) != the_nodes);
		freenode(node);
	}
	if((elmt = the_elmts) != (ELMTPTR)NULL){
		elmt = elmt->next;
		do {
			freeelmt(elmt);
		} while((elmt = elmt->next) != the_elmts);
		freeelmt(elmt);
	}
	the_nodes = (NODEPTR)NULL;
	the_elmts = (ELMTPTR)NULL;
}

freeelmt(elmt)
/*
	Frees a elmt and removes it from the circular list.
*/
ELMTPTR elmt;
{
	if(elmt == the_elmts){
		the_elmts = elmt->next;
		if(the_elmts->next == the_elmts)
			the_elmts = (ELMTPTR)NULL;
	}
	elmt->next->last = elmt->last;
	elmt->last->next = elmt->next;
	if(ues > 0)
		sfree(uetype, elmt->user);
	sfree(etype, elmt);
}

freenode(node)
/*
	Frees a node and its attendent nodeneighs, and removes it 
	from the circular list.
*/
NODEPTR node;
{
	NODENEIGHPTR n, nstart, n1;
	ALIASPTR ali, alj, astart;
	n = nstart = node->n;
	if(n != (NODENEIGHPTR)NULL){
		do {
			n1 = n->next;
			sfree(nntype, n);
		} while((n = n1) != nstart);
	}
	if(node == the_nodes){
		the_nodes = node->next;
		if(the_nodes->next == the_nodes)
			the_nodes = (NODEPTR)NULL;
	}
	node->next->last = node->last;
	node->last->next = node->next;
	if(node->b != (BDYPTR)NULL)
		freeb(node->b);
	if(uns > 0)
		sfree(untype, node->user);
	sfree(ntype, node);
}

freeb(bs)
BDYPTR bs;
/*
	Frees a boundary structure and any aliases that come along with it.
*/
{
	ALIASPTR ali, astart, alj;
	if(bs->alias != (ALIASPTR)NULL){
		ali = astart = bs->alias;
		if(astart == (ALIASPTR)NULL)
			abort(108);   /* Should have an alias but doesn't */
		do {
			alj = ali->next;
			sfree(atype, ali);
		} while((ali = alj) != astart);
	}
	if(bs->oppalias != (ALIASPTR)NULL)
		sfree(atype, bs->oppalias);
	if(ubs > 0)
		sfree(ubtype, bs->user);
	sfree(btype, bs);
}


/* The following code was originally in smalloc.c */

#define MINSIZE		8
#define MAXSIZE		400
#define BLOCKSIZE	4000
#define BITSPERINT	32
#define MAXINTSPERBLOCK	(BLOCKSIZE/(BITSPERINT*MINSIZE) + 1)
#define BNULL		(BLOCKPTR)0
#define MAXTYPES	12

extern char *malloc();

typedef struct block {
	unsigned int vacant[MAXINTSPERBLOCK];
	struct block *next;
/* 
MAXINTSPERBLOCK is 16 so this declaration makes number of ints
before data 18 rather than 17
*/
int align_dataBLOCK;
	char data[BLOCKSIZE];
} BLOCK, *BLOCKPTR;

BLOCKPTR block_alloc();
BLOCKPTR	curr_block,	known_blk_ptrs[MAXTYPES];;
int		curr_iv,	iv[MAXTYPES];
int		curr_ib,	ib[MAXTYPES];
int		curr_size,	known_sizes[MAXTYPES];
int		curr_flags,	known_flags[MAXTYPES];
int		knowntypes = 0;

unsigned int mask[BITSPERINT];

int init_smalloc(size)
int size;
{
	int ibs;

#ifdef EBUG
printf("size = %d\t", size);
#endif
/* Change size to be a multiple of 8 (if it isn't already) */
	if (size%8) size = size + (8 - (size%8));
#ifdef EBUG
printf("-> %d<<<\n", size);
#endif

	if(size < MINSIZE){
		fprintf(stderr, 
		"smalloc: size is %d: less than minimum size %d\n",
		size, MINSIZE);
		abort(8005);
	}
	for(ibs=0; ibs<BITSPERINT; ibs++)
		mask[ibs] = 1<<ibs;
	curr_size = known_sizes[knowntypes] = size;
	curr_block = BNULL;
	iv[knowntypes] = ib[knowntypes] = 0;
	if(size < MAXSIZE){
		curr_flags = known_flags[knowntypes] = BLOCKSIZE/(BITSPERINT*size) + 1;
		if((known_blk_ptrs[knowntypes] = block_alloc()) == BNULL)
			return(-1);
	} else {
		known_flags[knowntypes] = 0;
	}
	knowntypes++;
	return(knowntypes-1);
}

char *smalloc(type)
int type;
{
	BLOCKPTR start, block_alloc();
	int ibs, ivs;
	char *ptr;

	curr_size = known_sizes[type];

	if(curr_size >= MAXSIZE){
#ifdef EBUG
		printf("malloc %d\n", curr_size);
#endif
		ptr = malloc(curr_size);
		if(ptr == NULL){
			fprintf(stderr, "malloc failed: size %d", curr_size);
			abort(999);
		}
		known_flags[type]++;
		return(ptr);
	}
	curr_block = start =	known_blk_ptrs[type];
	curr_iv =		iv[type];
	curr_ib =		ib[type];
	curr_flags =		known_flags[type];

	do {
		ivs = curr_iv;
		do {
			if(curr_block->vacant[curr_iv] != 0){
				ibs = curr_ib;
				do {
					if((curr_block->vacant[curr_iv]) & (mask[curr_ib]))
						goto foundit;
				} while((curr_ib = (curr_ib+1)%BITSPERINT) != ibs);
				fprintf(stderr, "error in smalloc: size %d", curr_size);
				abort(801);
			}
		} while((curr_iv = (curr_iv+1)%curr_flags) != ivs);
	} while((curr_block = curr_block->next) != start);

	curr_block = block_alloc();

foundit:
	iv[type] = curr_iv;
	ib[type] = curr_ib;
	curr_block->vacant[curr_iv] &= ~mask[curr_ib];
	ptr = curr_block->data + curr_size*(BITSPERINT*curr_iv + curr_ib);
	return(ptr);
}

BLOCKPTR block_alloc()
{
	BLOCKPTR new;
	int iv, ib;

#ifdef EBUG
	printf("malloc %d\n", sizeof(BLOCK));
#endif
	new = (BLOCKPTR)malloc(sizeof(BLOCK));
	if(new == BNULL){
#ifdef CUBIX
		fprintf(stdsys, "block_alloc failed %d bytes in use\n", s_inventory());
#endif
		abort(s_inventory()/1000);
	}

	for(iv=0; iv<curr_flags; iv++){
		new->vacant[iv] = 0;
		for(ib=0; ib < BITSPERINT; ib++){
			if((iv*BITSPERINT + ib + 1)*curr_size > BLOCKSIZE)
				continue;
			new->vacant[iv] |= mask[ib];
		}
	}

	if(curr_block == BNULL){
		curr_block = new;
		curr_block->next = curr_block;
	} else {
		new->next = curr_block->next;
		curr_block->next = new;
	}
	curr_iv = curr_ib = 0;
	return(new);
}

sfree(type, ptr)
int type;
char *ptr;
{
	BLOCKPTR start;
	int iv, ib, posit;
	curr_block = start = known_blk_ptrs[type];
	curr_size = known_sizes[type];
	if(curr_size >= MAXSIZE){
		free(ptr);
		known_flags[type]--;
		return;
	}
	if(start == BNULL)
		goto fail;
	do {
		posit = ptr - curr_block->data;
		if(posit >= 0 && posit < BLOCKSIZE){
			posit /= curr_size;
			iv = posit/BITSPERINT;
			ib = posit%BITSPERINT;
			if(curr_block->vacant[iv] & mask[ib]){
				abort(802);
			}
			curr_block->vacant[iv] |= mask[ib];
			return;
		}
	} while((curr_block = curr_block->next) != start);
fail:
#ifdef CUBIX
	fprintf(stdsys, "sfree: not mine: type %d size %d\n", type, curr_size);
#endif
	abort(803);
}

int s_inventory()
{
	
	BLOCKPTR start, loop;
	int type, nblock = 0, mallocked = 0;
	for(type=0; type<knowntypes; type++){
		if(known_sizes[type] >= MAXSIZE){
			mallocked += known_sizes[type] * known_flags[type];
		} else {
			start = loop = known_blk_ptrs[type];
			do {
				nblock++;
			} while((loop = loop->next) != start);
		}
	}
	return(nblock*sizeof(BLOCK) + mallocked);
}

